@@ -67,6 +67,8 @@ use structopt::{StructOpt, clap::AppSettings};
 use crate::{
     ArmorLabel,
     EncryptAs,
+    InlineSignAs,
+    Password,
     SignAs,
 };
 
@@ -101,6 +103,9 @@ enum Operation {
         /// Don't ASCII-armor output.
         #[structopt(long)]
         no_armor: bool,
+        /// Protect the newly generated key with the given password.
+        #[structopt(long)]
+        with_key_password: Option<String>,
         /// UserIDs for the generated key.
         userids: Vec<String>,
     },
@@ -123,6 +128,9 @@ enum Operation {
         /// Emit the digest algorithm used to the specified file.
         #[structopt(long)]
         micalg_out: Option<String>,
+        /// Try to decrypt the signing KEYS with these passwords.
+        #[structopt(long, number_of_values = 1)]
+        with_key_password: Vec<String>,
         /// Keys for signing.
         keys: Vec<String>,
     },
@@ -146,7 +154,7 @@ enum Operation {
         /// Don't ASCII-armor output.
         #[structopt(long)]
         no_armor: bool,
-        /// Encrypt binary data, UTF-8 text, or MIME data.
+        /// Encrypt binary data or UTF-8 text.
         #[structopt(default_value = "binary", long = "as")]
         as_: EncryptAs,
         /// Encrypt with passwords.
@@ -155,6 +163,9 @@ enum Operation {
         /// Keys for signing.
         #[structopt(long, number_of_values = 1)]
         sign_with: Vec<String>,
+        /// Try to decrypt the signing KEYS with these passwords.
+        #[structopt(long, number_of_values = 1)]
+        with_key_password: Vec<String>,
         /// Encrypt for these certs.
         certs: Vec<String>,
     },
@@ -182,6 +193,9 @@ enum Operation {
         /// Consider signatures after this date invalid.
         #[structopt(long, parse(try_from_str = parse_bound_round_up))]
         verify_not_after: Option<DateTime<Utc>>,
+        /// Try to decrypt the encryption KEYS with these passwords.
+        #[structopt(long, number_of_values = 1)]
+        with_key_password: Vec<String>,
         /// Try to decrypt with these keys.
         keys: Vec<String>,
     },
@@ -196,6 +210,47 @@ enum Operation {
     #[structopt(display_order = 900)]
     Dearmor {
     },
+    /// Splits Signatures from an Inline-Signed Message.
+    #[structopt(display_order = 950)]
+    InlineDetach {
+        /// Don't ASCII-armor the signatures.
+        #[structopt(long)]
+        no_armor: bool,
+        /// Write Signatures here.
+        #[structopt(long)]
+        signatures_out: String,
+    },
+    /// Verifies Inline-Signed Messages.
+    #[structopt(display_order = 951)]
+    InlineVerify {
+        /// Consider signatures before this date invalid.
+        #[structopt(long, parse(try_from_str = parse_bound_round_down))]
+        not_before: Option<DateTime<Utc>>,
+        /// Consider signatures after this date invalid.
+        #[structopt(long, parse(try_from_str = parse_bound_round_up))]
+        not_after: Option<DateTime<Utc>>,
+        /// Write verification result here.
+        #[structopt(long)]
+        verifications_out: Option<String>,
+        /// Certs for verification.
+        certs: Vec<String>,
+    },
+    /// Creates Inline-Signed Messages.
+    #[structopt(display_order = 951)]
+    InlineSign {
+        /// Don't ASCII-armor output.
+        #[structopt(long)]
+        no_armor: bool,
+        /// Sign binary data, UTF-8 text, or using the Cleartext
+        /// Signature Framework.
+        #[structopt(default_value = "binary", long = "as")]
+        as_: InlineSignAs,
+        /// Try to decrypt the signing KEYS with these passwords.
+        #[structopt(long, number_of_values = 1)]
+        with_key_password: Vec<String>,
+        /// Keys for signing.
+        keys: Vec<String>,
+    },
     /// Unsupported subcommand.
     #[structopt(external_subcommand)]
     Unsupported(Vec<String>),
@@ -271,11 +326,14 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             }
         },
 
-        Operation::GenerateKey { no_armor, userids, } => {
+        Operation::GenerateKey { no_armor, with_key_password, userids, } => {
             let mut op = sop.generate_key()?;
             if no_armor {
                 op = op.no_armor();
             }
+            if let Some(p) = with_key_password {
+                op = op.with_key_password(get_password(p)?)?;
+            }
             for u in userids {
                 op = op.userid(&u);
             }
@@ -290,11 +348,20 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             op.keys(&mut io::stdin())?.write_to(&mut io::stdout())?;
         },
 
-        Operation::Sign { no_armor, as_, micalg_out, keys, } => {
+        Operation::Sign {
+            no_armor,
+            as_,
+            micalg_out,
+            with_key_password,
+            keys,
+        } => {
             let mut op = sop.sign()?.mode(as_);
             if no_armor {
                 op = op.no_armor();
             }
+            for p in with_key_password {
+                op = op.with_key_password(get_password_unchecked(p)?)?;
+            }
             for mut key in load_files(keys)? {
                 op = op.key(&mut key)?;
             }
@@ -325,19 +392,27 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             }
         },
 
-        Operation::Encrypt { no_armor, as_, with_password, sign_with, certs, } =>
+        Operation::Encrypt {
+            no_armor,
+            as_,
+            with_password,
+            with_key_password,
+            sign_with,
+            certs,
+        } =>
         {
             let mut op = sop.encrypt()?.mode(as_);
             if no_armor {
                 op = op.no_armor();
             }
+            for p in with_key_password {
+                op = op.with_key_password(get_password_unchecked(p)?)?;
+            }
             for mut key in load_files(sign_with)? {
                 op = op.sign_with_key(&mut key)?;
             }
-            for mut stream in load_files(with_password)? {
-                let mut pw = String::new();
-                stream.read_to_string(&mut pw)?;
-                op = op.with_password(pw.into())?;
+            for pw in with_password.into_iter().map(get_password) {
+                op = op.with_password(pw?)?;
             }
             for mut cert in load_files(certs)? {
                 op = op.with_cert(&mut cert)?;
@@ -353,6 +428,7 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             verify_with,
             verify_not_before,
             verify_not_after,
+            with_key_password,
             keys,
         } => {
             let session_key_out: Option<Box<dyn io::Write>> =
@@ -383,10 +459,12 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
                 op = op.with_session_key(sk.parse()?)?;
             }
 
-            for mut stream in load_files(with_password)? {
-                let mut pw = String::new();
-                stream.read_to_string(&mut pw)?;
-                op = op.with_password(pw.into())?;
+            for pw in with_password.into_iter().map(get_password_unchecked) {
+                op = op.with_password(pw?)?;
+            }
+
+            for p in with_key_password {
+                op = op.with_key_password(get_password_unchecked(p)?)?;
             }
 
             for mut key in load_files(keys)? {
@@ -432,6 +510,62 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
                 .write_to(&mut io::stdout())?;
         },
 
+        Operation::InlineDetach { no_armor, signatures_out, } => {
+            let mut op = sop.inline_detach()?;
+            if no_armor {
+                op = op.no_armor();
+            }
+            let mut signatures_out = create_file(signatures_out)?;
+            let signatures =
+                op.message(&mut io::stdin())?.write_to(&mut io::stdout())?;
+            signatures_out.write_all(&signatures)?;
+        },
+
+        Operation::InlineVerify {
+            not_before,
+            not_after,
+            verifications_out,
+            certs,
+        } => {
+            let mut op = sop.inline_verify()?;
+
+            if let Some(t) = not_before {
+                op = op.not_before(t.into());
+            }
+            if let Some(t) = not_after {
+                op = op.not_after(t.into());
+            }
+            let mut verifications_out: Box<dyn io::Write> =
+                if let Some(f) = verifications_out {
+                    Box::new(create_file(f)?)
+                } else {
+                    Box::new(io::sink())
+                };
+            for mut cert in load_files(certs)? {
+                op = op.certs(&mut cert)?;
+            }
+            let verifications =
+                op.message(&mut io::stdin())?.write_to(&mut io::stdout())?;
+
+            for v in verifications {
+                writeln!(&mut verifications_out, "{}", v)?;
+            }
+        },
+
+        Operation::InlineSign { no_armor, as_, with_key_password, keys, } => {
+            let mut op = sop.inline_sign()?.mode(as_);
+            if no_armor {
+                op = op.no_armor();
+            }
+            for p in with_key_password {
+                op = op.with_key_password(get_password_unchecked(p)?)?;
+            }
+            for mut key in load_files(keys)? {
+                op = op.key(&mut key)?;
+            }
+            op.data(&mut io::stdin())?.write_to(&mut io::stdout())?;
+        },
+
         Operation::Unsupported(args) => {
             return Err(anyhow::Error::from(Error::UnsupportedSubcommand))
                 .context(format!("Subcommand {} is not supported", args[0]));
@@ -503,6 +637,32 @@ fn load_file<S: AsRef<str>>(file: S)
         .context(format!("Failed to open file {:?}", f))
 }
 
+/// Gets the password from the given location.
+///
+/// Use this for passwords given to create artifacts.
+///
+/// Passwords are indirect data types, meaning that they are
+/// loaded from files or special designators.
+fn get_password<S: AsRef<str>>(location: S) -> Result<Password> {
+    let mut stream = load_file(location)?;
+    let mut pw = Vec::new();
+    stream.read_to_end(&mut pw)?;
+    Ok(Password::new(pw)?)
+}
+
+/// Gets the password from the given location.
+///
+/// Use this for passwords given to consume artifacts.
+///
+/// Passwords are indirect data types, meaning that they are
+/// loaded from files or special designators.
+fn get_password_unchecked<S: AsRef<str>>(location: S) -> Result<Password> {
+    let mut stream = load_file(location)?;
+    let mut pw = Vec::new();
+    stream.read_to_end(&mut pw)?;
+    Ok(Password::new_unchecked(pw))
+}
+
 /// Creates the given (special) file.
 fn create_file<S: AsRef<str>>(file: S) -> Result<std::fs::File> {
     let f = file.as_ref();
@@ -700,6 +860,10 @@ enum Error {
     #[error("A indirect input parameter is a special designator matches file")]
     AmbiguousInput,
 
+    /// Options were supplied that are incompatible with each other.
+    #[error("Options were supplied that are incompatible with each other")]
+    IncompatibleOptions,
+
     /// An I/O operation failed.
     #[error("I/O operation failed")]
     IoError(#[source] io::Error),
@@ -725,6 +889,7 @@ impl From<crate::Error> for Error {
             E::MissingInput => CE::MissingInput,
             E::KeyIsProtected => CE::KeyIsProtected,
             E::AmbiguousInput => CE::AmbiguousInput,
+            E::IncompatibleOptions => CE::IncompatibleOptions,
             E::NotImplemented => CE::UnsupportedSubcommand,
             E::IoError(e) => CE::IoError(e),
         }
@@ -752,6 +917,7 @@ impl From<Error> for i32 {
             UnsupportedSpecialPrefix => 71,
             AmbiguousInput => 73,
             KeyCannotSign => 79,
+            IncompatibleOptions => 83,
             IoError(_) => 1,
         }
     }
