@@ -1,4 +1,9 @@
 //! A Rust implementation of the Stateless OpenPGP Protocol.
+//!
+//! This interface is the Rust equivalent of the [draft 04 of the
+//! Stateless OpenPGP Command Line Interface].
+//!
+//! [draft 04 of the Stateless OpenPGP Command Line Interface]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html
 
 use std::{
     fmt,
@@ -53,6 +58,21 @@ pub trait SOP {
     ///
     /// Customize the operation using the builder [`Dearmor`].
     fn dearmor<'a>(&'a self) -> Result<Box<dyn Dearmor + 'a>>;
+
+    /// Splits Signatures from an Inline-Signed Message.
+    ///
+    /// Customize the operation using the builder [`InlineDetach`].
+    fn inline_detach<'a>(&'a self) -> Result<Box<dyn InlineDetach + 'a>>;
+
+    /// Verifies an Inline-Signed Message.
+    ///
+    /// Customize the operation using the builder [`InlineVerify`].
+    fn inline_verify<'a>(&'a self) -> Result<Box<dyn InlineVerify + 'a>>;
+
+    /// Creates an Inline-Signed Message.
+    ///
+    /// Customize the operation using the builder [`InlineSign`].
+    fn inline_sign<'a>(&'a self) -> Result<Box<dyn InlineSign + 'a>>;
 }
 
 pub trait Version<'a> {
@@ -89,6 +109,10 @@ pub trait GenerateKey<'a> {
     /// Disables armor encoding.
     fn no_armor(self: Box<Self>) -> Box<dyn GenerateKey<'a> + 'a>;
 
+    /// Protects the newly generated key with the given password.
+    fn with_key_password(self: Box<Self>, password: Password)
+                         -> Result<Box<dyn GenerateKey<'a> + 'a>>;
+
     /// Adds a User ID.
     fn userid(self: Box<Self>, userid: &str) -> Box<dyn GenerateKey<'a> + 'a>;
 
@@ -128,6 +152,13 @@ pub trait Sign<'a> {
     fn keys(self: Box<Self>, keys: &mut (dyn io::Read + Send + Sync))
             -> Result<Box<dyn Sign<'a> + 'a>>;
 
+    /// Adds a password to unlock the signing keys with.
+    ///
+    /// All supplied passwords will be used to try to unlock all
+    /// signing keys.
+    fn with_key_password(self: Box<Self>, password: Password)
+                         -> Result<Box<dyn Sign<'a> + 'a>>;
+
     /// Signs data.
     fn data(self: Box<Self>, data: &'a mut (dyn io::Read + Send + Sync))
             -> Result<Box<dyn ReadyWithResult<Micalg> + 'a>>;
@@ -259,6 +290,13 @@ pub trait Encrypt<'a> {
     fn sign_with_keys(self: Box<Self>, keys: &mut (dyn io::Read + Send + Sync))
                       -> Result<Box<dyn Encrypt<'a> + 'a>>;
 
+    /// Adds a password to unlock the signing keys with.
+    ///
+    /// All supplied passwords will be used to try to unlock all
+    /// signing keys.
+    fn with_key_password(self: Box<Self>, password: Password)
+                         -> Result<Box<dyn Encrypt<'a> + 'a>>;
+
     /// Encrypts with the given password.
     fn with_password(self: Box<Self>, password: Password)
                      -> Result<Box<dyn Encrypt<'a> + 'a>>;
@@ -319,6 +357,12 @@ pub trait Decrypt<'a> {
     fn with_keys(self: Box<Self>, key: &mut (dyn io::Read + Send + Sync))
                  -> Result<Box<dyn Decrypt<'a> + 'a>>;
 
+    /// Adds a password to unlock the decryption keys with.
+    ///
+    /// All supplied passwords will be used to try to unlock all keys.
+    fn with_key_password(self: Box<Self>, password: Password)
+                         -> Result<Box<dyn Decrypt<'a> + 'a>>;
+
     /// Decrypts `ciphertext`, returning verification results and
     /// plaintext.
     fn ciphertext(self: Box<Self>,
@@ -344,43 +388,156 @@ pub trait Dearmor<'a> {
         -> Result<Box<dyn Ready + 'a>>;
 }
 
-/// A normalized password.
+/// Builder for [`SOP::inline_detach`].
+pub trait InlineDetach<'a> {
+    /// Disables armor encoding.
+    fn no_armor(self: Box<Self>) -> Box<dyn InlineDetach<'a> + 'a>;
+
+    /// Splits Signatures from the Inline-Signed Message.
+    fn message(self: Box<Self>,
+               data: &'a mut (dyn io::Read + Send + Sync))
+               -> Result<Box<dyn ReadyWithResult<Vec<u8>> + 'a>>;
+}
+
+/// Builder for [`SOP::inline_verify`].
+pub trait InlineVerify<'a> {
+    /// Makes SOP consider signatures before this date invalid.
+    fn not_before(self: Box<Self>, t: SystemTime) -> Box<dyn InlineVerify<'a> + 'a>;
+
+    /// Makes SOP consider signatures after this date invalid.
+    fn not_after(self: Box<Self>, t: SystemTime) -> Box<dyn InlineVerify<'a> + 'a>;
+
+    /// Adds the verification cert.
+    fn cert(self: Box<Self>, cert: &mut (dyn io::Read + Send + Sync))
+            -> Result<Box<dyn InlineVerify<'a> + 'a>>;
+
+    /// Adds the verification certs.
+    ///
+    /// Like [`Verify::cert`], but for multiple certs.
+    fn certs(self: Box<Self>, certs: &mut (dyn io::Read + Send + Sync))
+             -> Result<Box<dyn InlineVerify<'a> + 'a>>;
+
+    /// Verifies an Inline-Signed Message.
+    fn message(self: Box<Self>,
+               data: &'a mut (dyn io::Read + Send + Sync))
+               -> Result<Box<dyn ReadyWithResult<Vec<Verification>> + 'a>>;
+}
+
+/// Builder for [`SOP::inline_sign`].
+pub trait InlineSign<'a> {
+    /// Disables armor encoding.
+    fn no_armor(self: Box<Self>) -> Box<dyn InlineSign<'a> + 'a>;
+
+    /// Sets signature mode.
+    fn mode(self: Box<Self>, mode: InlineSignAs) -> Box<dyn InlineSign<'a> + 'a>;
+
+    /// Adds the signer key.
+    fn key(self: Box<Self>, key: &mut (dyn io::Read + Send + Sync))
+           -> Result<Box<dyn InlineSign<'a> + 'a>>;
+
+    /// Adds the signer keys.
+    ///
+    /// Like [`Sign::key`], but for multiple keys.
+    fn keys(self: Box<Self>, keys: &mut (dyn io::Read + Send + Sync))
+            -> Result<Box<dyn InlineSign<'a> + 'a>>;
+
+    /// Adds a password to unlock the signing keys with.
+    ///
+    /// All supplied passwords will be used to try to unlock all
+    /// signing keys.
+    fn with_key_password(self: Box<Self>, password: Password)
+                         -> Result<Box<dyn InlineSign<'a> + 'a>>;
+
+    /// Signs data.
+    fn data(self: Box<Self>, data: &'a mut (dyn io::Read + Send + Sync))
+            -> Result<Box<dyn Ready + 'a>>;
+}
+
+/// A password.
 ///
 /// See [Passwords are Human-Readable] in the SOP spec.
 ///
-///   [Passwords are Human-Readable]: https://datatracker.ietf.org/doc/html/draft-dkg-openpgp-stateless-cli-02#section-7.8
+///   [Passwords are Human-Readable]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-passwords-are-human-readabl
 pub struct Password(Box<[u8]>);
 
-impl From<&str> for Password {
-    fn from(p: &str) -> Self {
-        p.to_string().into()
-    }
-}
-
-impl From<String> for Password {
-    fn from(p: String) -> Self {
-        Password(Self::normalize(p).into_bytes().into_boxed_slice())
-    }
-}
-
-impl AsRef<[u8]> for Password {
-    fn as_ref(&self) -> &[u8] {
-        &self.0
-    }
-}
-
 impl Password {
-    fn normalize(p: String) -> String {
-        // XXX: Maybe do additional checks.
-        let normalized = p.trim_end().to_string();
+    /// Returns a `Password` that is guaranteed to be human-readable.
+    ///
+    /// Use this function when you get a password from the user to
+    /// generate an artifact with (see [Generating Material with
+    /// Human-Readable Passwords]).
+    ///
+    /// [Generating Material with Human-Readable Passwords]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-generating-material-with-hu
+    pub fn new(password: Vec<u8>) -> Result<Password> {
+        let s = String::from_utf8(password)
+            .map_err(|_| Error::PasswordNotHumanReadable)?;
 
-        // Wipe original string.
-        let mut p = p.into_bytes();
-        unsafe {
-            memsec::memzero(p.as_mut_ptr(), p.len());
+        // Check for leading whitespace.
+        if s.trim_start().len() != s.len() {
+            return Err(Error::PasswordNotHumanReadable);
         }
 
-        normalized
+        // Check for odd whitespace.
+        if s.chars().any(|c| c.is_whitespace() && c != ' ') {
+            return Err(Error::PasswordNotHumanReadable);
+        }
+
+        // XXX: Check that the password is in Unicode Normal Form C,
+        // but I don't think that is possible with Rust's stdlib.
+
+        Ok(Password(s.into_bytes().into()))
+    }
+
+    /// Returns a `Password` without further checking.
+    ///
+    /// Use this function when you get a password from the user that
+    /// is used when consuming an artifact (see [Consuming
+    /// Password-protected Material]).
+    ///
+    /// [Consuming Password-protected Material]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-consuming-password-protecte
+    pub fn new_unchecked(password: Vec<u8>) -> Password {
+        Password(password.into())
+    }
+
+    /// Returns the normalized password.
+    ///
+    /// Use this function when you generate an artifact using a
+    /// password (see [Generating Material with Human-Readable
+    /// Passwords]).
+    ///
+    /// [Generating Material with Human-Readable Passwords]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-generating-material-with-hu
+    pub fn normalized(&self) -> &[u8] {
+        // First, let's hope it is UTF-8.
+        if let Ok(p) = std::str::from_utf8(&self.0) {
+            p.trim_end().as_bytes()
+        } else {
+            // As a best effort for now, drop ASCII-whitespace from
+            // the end.
+            let mut p = &self.0[..];
+            while ! p.is_empty() && p[p.len() - 1].is_ascii_whitespace() {
+                p = &p[..p.len() - 1];
+            }
+            p
+        }
+    }
+
+    /// Returns the password in all possible variants.
+    ///
+    /// Use this function to try all possible variants
+    /// (i.e. normalized, as-is, ..) when consuming an artifact (see
+    /// [Consuming Password-protected Material]).
+    ///
+    /// [Consuming Password-protected Material]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-consuming-password-protecte
+    pub fn variants(&self) -> impl Iterator<Item = &[u8]> {
+        std::iter::once(self.normalized())
+            .filter_map(move |normalized| {
+                if normalized.len() < self.0.len() {
+                    Some(normalized)
+                } else {
+                    None
+                }
+            })
+            .chain(std::iter::once(&self.0[..]))
     }
 }
 
@@ -561,7 +718,7 @@ fn from_hex(hex: &str, pretty: bool) -> ParseResult<Vec<u8>> {
 /// This is used by [`SOP::sign`] to select the signature type.  See
 /// [`sop sign`].
 ///
-///   [`sop sign`]: https://datatracker.ietf.org/doc/html/draft-dkg-openpgp-stateless-cli-02#section-3.4
+///   [`sop sign`]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-sign-create-detached-signat
 #[derive(Clone, Copy, Debug, Hash, PartialEq, Eq)]
 pub enum SignAs {
     Binary,
@@ -579,10 +736,6 @@ impl From<EncryptAs> for SignAs {
         match a {
             EncryptAs::Binary => SignAs::Binary,
             EncryptAs::Text => SignAs::Text,
-            // XXX: We should inspect the serialized MIME structure
-            // and use Text if it is UTF-8, Binary otherwise.  But, we
-            // cannot be bothered at this point.
-            EncryptAs::MIME => SignAs::Binary,
         }
     }
 }
@@ -608,17 +761,58 @@ impl fmt::Display for SignAs {
     }
 }
 
+/// Inline Signature type.
+///
+/// This is used by [`SOP::inline_sign`] to select the signature type.
+/// See [`sop inline-sign`].
+///
+///   [`sop inline-sign`]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-inline-sign-create-an-inlin
+#[derive(Clone, Copy, Debug, Hash, PartialEq, Eq)]
+pub enum InlineSignAs {
+    Binary,
+    Text,
+    ClearSigned,
+}
+
+impl Default for InlineSignAs {
+    fn default() -> Self {
+        InlineSignAs::Binary
+    }
+}
+
+impl std::str::FromStr for InlineSignAs {
+    type Err = ParseError;
+    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
+        match s {
+            "binary" => Ok(InlineSignAs::Binary),
+            "text" => Ok(InlineSignAs::Text),
+            "clearsigned" => Ok(InlineSignAs::ClearSigned),
+            _ => Err(ParseError(format!(
+                "{:?}, expected one of {{binary|text|clearsigned}}", s))),
+        }
+    }
+}
+
+impl fmt::Display for InlineSignAs {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            InlineSignAs::Binary => f.write_str("binary"),
+            InlineSignAs::Text => f.write_str("text"),
+            InlineSignAs::ClearSigned => f.write_str("clearsigned"),
+        }
+    }
+}
+
 /// Plaintext data format.
 ///
 /// This is used by [`SOP::encrypt`] to select the data format.  See
 /// [`sop encrypt`].
 ///
-///   [`sop encrypt`]: https://datatracker.ietf.org/doc/html/draft-dkg-openpgp-stateless-cli-02#section-3.6
+///   [`sop encrypt`]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-encrypt-encrypt-a-message
 #[derive(Clone, Copy, Debug, Hash, PartialEq, Eq)]
 pub enum EncryptAs {
     Binary,
     Text,
-    MIME,
 }
 
 impl Default for EncryptAs {
@@ -633,9 +827,8 @@ impl std::str::FromStr for EncryptAs {
         match s {
             "binary" => Ok(EncryptAs::Binary),
             "text" => Ok(EncryptAs::Text),
-            "mime" => Ok(EncryptAs::MIME),
             _ => Err(ParseError(format!(
-                "{}, expected one of {{binary|text|mime}}", s))),
+                "{}, expected one of {{binary|text}}", s))),
         }
     }
 }
@@ -645,7 +838,6 @@ impl fmt::Display for EncryptAs {
         match self {
             EncryptAs::Binary => f.write_str("binary"),
             EncryptAs::Text => f.write_str("text"),
-            EncryptAs::MIME => f.write_str("mime"),
         }
     }
 }
@@ -655,7 +847,7 @@ impl fmt::Display for EncryptAs {
 /// This is used by [`SOP::armor`] to control the framing that is
 /// emitted.  See [`sop armor`].
 ///
-///   [`sop armor`]: https://datatracker.ietf.org/doc/html/draft-dkg-openpgp-stateless-cli-02#section-3.8
+///   [`sop armor`]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-armor-convert-binary-to-asc
 #[derive(Clone, Copy, Debug, Hash, PartialEq, Eq)]
 pub enum ArmorLabel {
     Auto,
@@ -764,7 +956,7 @@ pub type Result<T> = std::result::Result<T, Error>;
 ///
 /// These are the errors [defined] by the Stateless OpenPGP Protocol.
 ///
-///   [defined]: https://datatracker.ietf.org/doc/html/draft-dkg-openpgp-stateless-cli-02#section-6
+///   [defined]: https://www.ietf.org/archive/id/draft-dkg-openpgp-stateless-cli-04.html#name-failure-modes
 #[derive(thiserror::Error, Debug)]
 pub enum Error {
     /// No acceptable signatures found ("sop verify").
@@ -823,8 +1015,8 @@ pub enum Error {
     MissingInput,
 
     /// A "KEY" input is protected (locked) with a password, and "sop" cannot
-    /// unlock it.
-    #[error("A KEY input is protected with a password")]
+    /// unlock it with any of the --with-key-password options.
+    #[error("A KEY input is protected with a password and unlocking failed")]
     KeyIsProtected,
 
     /// A indirect input parameter is a special designator (it starts with
@@ -832,6 +1024,10 @@ pub enum Error {
     #[error("A indirect input parameter is a special designator matches file")]
     AmbiguousInput,
 
+    /// Options were supplied that are incompatible with each other.
+    #[error("Options were supplied that are incompatible with each other")]
+    IncompatibleOptions,
+
     /// Operation not implemented.
     #[error("Operation not implemented")]
     NotImplemented,
@@ -888,7 +1084,7 @@ mod tests {
     #[test]
     fn encrypt_as_roundtrip() -> Result<()> {
         use EncryptAs::*;
-        for a in &[Text, Binary, MIME] {
+        for a in &[Text, Binary] {
             let s = a.to_string();
             let b: EncryptAs = s.parse().unwrap();
             assert_eq!(a, &b);
